from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import OrderedDict
import warnings
import logging
import os
import re
import subprocess

import numpy as np
from six.moves import xrange

known_number_types = (int, float, np.float16, np.float32, np.float64,
                      np.int8, np.int16, np.int32, np.int32, np.int64,
                      np.uint8, np.uint16, np.uint32, np.uint64)


CLEVERHANS_ROOT = os.path.dirname(os.path.dirname(__file__))


class _ArgsWrapper(object):

  """
  Wrapper that allows attribute access to dictionaries
  """

  def __init__(self, args):
    if not isinstance(args, dict):
      args = vars(args)
    self.args = args

  def __getattr__(self, name):
    return self.args.get(name)


class AccuracyReport(object):

  """
  An object summarizing the accuracy results for experiments involving
  training on clean examples or adversarial examples, then evaluating
  on clean or adversarial examples.
  """

  def __init__(self):
    self.clean_train_clean_eval = 0.
    self.clean_train_adv_eval = 0.
    self.adv_train_clean_eval = 0.
    self.adv_train_adv_eval = 0.

    # Training data accuracy results to be used by tutorials
    self.train_clean_train_clean_eval = 0.
    self.train_clean_train_adv_eval = 0.
    self.train_adv_train_clean_eval = 0.
    self.train_adv_train_adv_eval = 0.


def batch_indices(batch_nb, data_length, batch_size):
  """
  This helper function computes a batch start and end index
  :param batch_nb: the batch number
  :param data_length: the total length of the data being parsed by batches
  :param batch_size: the number of inputs in each batch
  :return: pair of (start, end) indices
  """
  # Batch start and end index
  start = int(batch_nb * batch_size)
  end = int((batch_nb + 1) * batch_size)

  # When there are not enough inputs left, we reuse some to complete the
  # batch
  if end > data_length:
    shift = end - data_length
    start -= shift
    end -= shift

  return start, end


def other_classes(nb_classes, class_ind):
  """
  Returns a list of class indices excluding the class indexed by class_ind
  :param nb_classes: number of classes in the task
  :param class_ind: the class index to be omitted
  :return: list of class indices excluding the class indexed by class_ind
  """
  if class_ind < 0 or class_ind >= nb_classes:
    error_str = "class_ind must be within the range (0, nb_classes - 1)"
    raise ValueError(error_str)

  other_classes_list = list(range(nb_classes))
  other_classes_list.remove(class_ind)

  return other_classes_list


def to_categorical(y, nb_classes, num_classes=None):
  """
  Converts a class vector (integers) to binary class matrix.
  This is adapted from the Keras function with the same name.
  :param y: class vector to be converted into a matrix
            (integers from 0 to nb_classes).
  :param nb_classes: nb_classes: total number of classes.
  :param num_classses: depricated version of nb_classes
  :return: A binary matrix representation of the input.
  """
  if num_classes is not None:
    if nb_classes is not None:
      raise ValueError("Should not specify both nb_classes and its deprecated "
                       "alias, num_classes")
    warnings.warn("`num_classes` is deprecated. Switch to `nb_classes`."
                  " `num_classes` may be removed on or after 2019-04-23.")
    nb_classes = num_classes
    del num_classes
  y = np.array(y, dtype='int').ravel()
  n = y.shape[0]
  categorical = np.zeros((n, nb_classes))
  categorical[np.arange(n), y] = 1
  return categorical


def random_targets(gt, nb_classes):
  """
  Take in an array of correct labels and randomly select a different label
  for each label in the array. This is typically used to randomly select a
  target class in targeted adversarial examples attacks (i.e., when the
  search algorithm takes in both a source class and target class to compute
  the adversarial example).
  :param gt: the ground truth (correct) labels. They can be provided as a
             1D vector or 2D array of one-hot encoded labels.
  :param nb_classes: The number of classes for this task. The random class
                     will be chosen between 0 and nb_classes such that it
                     is different from the correct class.
  :return: A numpy array holding the randomly-selected target classes
           encoded as one-hot labels.
  """
  # If the ground truth labels are encoded as one-hot, convert to labels.
  if len(gt.shape) == 2:
    gt = np.argmax(gt, axis=1)

  # This vector will hold the randomly selected labels.
  result = np.zeros(gt.shape, dtype=np.int32)

  for class_ind in xrange(nb_classes):
    # Compute all indices in that class.
    in_cl = gt == class_ind
    size = np.sum(in_cl)

    # Compute the set of potential targets for this class.
    potential_targets = other_classes(nb_classes, class_ind)

    # Draw with replacement random targets among the potential targets.
    result[in_cl] = np.random.choice(potential_targets, size=size)

  # Encode vector of random labels as one-hot labels.
  result = to_categorical(result, nb_classes)
  result = result.astype(np.int32)

  return result


def pair_visual(*args, **kwargs):
  warnings.warn("`pair_visual` has moved to `cleverhans.plot.pyplot_image`. "
                "cleverhans.utils.pair_visual may be removed on or after "
                "2019-04-24.")
  from cleverhans.plot.pyplot_image import pair_visual as new_pair_visual
  return new_pair_visual(*args, **kwargs)


def grid_visual(*args, **kwargs):
  warnings.warn("`grid_visual` has moved to `cleverhans.plot.pyplot_image`. "
                "cleverhans.utils.grid_visual may be removed on or after "
                "2019-04-24.")
  from cleverhans.plot.pyplot_image import grid_visual as new_grid_visual
  return new_grid_visual(*args, **kwargs)


def get_logits_over_interval(*args, **kwargs):
  warnings.warn("`get_logits_over_interval` has moved to "
                "`cleverhans.plot.pyplot_image`. "
                "cleverhans.utils.get_logits_over_interval may be removed on "
                "or after 2019-04-24.")
  # pylint:disable=line-too-long
  from cleverhans.plot.pyplot_image import get_logits_over_interval as new_get_logits_over_interval
  return new_get_logits_over_interval(*args, **kwargs)


def linear_extrapolation_plot(*args, **kwargs):
  warnings.warn("`linear_extrapolation_plot` has moved to "
                "`cleverhans.plot.pyplot_image`. "
                "cleverhans.utils.linear_extrapolation_plot may be removed on "
                "or after 2019-04-24.")
  # pylint:disable=line-too-long
  from cleverhans.plot.pyplot_image import linear_extrapolation_plot as new_linear_extrapolation_plot
  return new_linear_extrapolation_plot(*args, **kwargs)


def set_log_level(level, name="cleverhans"):
  """
  Sets the threshold for the cleverhans logger to level
  :param level: the logger threshold. You can find values here:
                https://docs.python.org/2/library/logging.html#levels
  :param name: the name used for the cleverhans logger
  """
  logging.getLogger(name).setLevel(level)


def get_log_level(name="cleverhans"):
  """
  Gets the current threshold for the cleverhans logger
  :param name: the name used for the cleverhans logger
  """
  return logging.getLogger(name).getEffectiveLevel()


class TemporaryLogLevel(object):
  """
  A ContextManager that changes a log level temporarily.

  Note that the log level will be set back to its original value when
  the context manager exits, even if the log level has been changed
  again in the meantime.
  """

  def __init__(self, level, name):
    self.name = name
    self.level = level

  def __enter__(self):
    self.old_level = get_log_level(self.name)
    set_log_level(self.level, self.name)

  def __exit__(self, type, value, traceback):
    set_log_level(self.old_level, self.name)
    return True


def create_logger(name):
  """
  Create a logger object with the given name.

  If this is the first time that we call this method, then initialize the
  formatter.
  """
  base = logging.getLogger("cleverhans")
  if len(base.handlers) == 0:
    ch = logging.StreamHandler()
    formatter = logging.Formatter('[%(levelname)s %(asctime)s %(name)s] ' +
                                  '%(message)s')
    ch.setFormatter(formatter)
    base.addHandler(ch)

  return base


def deterministic_dict(normal_dict):
  """
  Returns a version of `normal_dict` whose iteration order is always the same
  """
  out = OrderedDict()
  for key in sorted(normal_dict.keys()):
    out[key] = normal_dict[key]
  return out


def ordered_union(l1, l2):
  """
  Return the union of l1 and l2, with a deterministic ordering.
  (Union of python sets does not necessarily have a consisten iteration
  order)
  :param l1: list of items
  :param l2: list of items
  :returns: list containing one copy of each item that is in l1 or in l2
  """
  out = []
  for e in l1 + l2:
    if e not in out:
      out.append(e)
  return out


def safe_zip(*args):
  """zip, with a guarantee that all arguments are the same length.
  (normal zip silently drops entries to make them the same length)
  """
  length = len(args[0])
  if not all(len(arg) == length for arg in args):
    raise ValueError("Lengths of arguments do not match: "
                     + str([len(arg) for arg in args]))
  return zip(*args)


def shell_call(command, **kwargs):
  """Calls shell command with argument substitution.

  Args:
    command: command represented as a list. Each element of the list is one
      token of the command. For example "cp a b" becomes ['cp', 'a', 'b']
      If any element of the list looks like '${NAME}' then it will be replaced
      by value from **kwargs with key 'NAME'.
    **kwargs: dictionary with argument substitution

  Returns:
    output of the command

  Raises:
    subprocess.CalledProcessError if command return value is not zero

  This function is useful when you need to do variable substitution prior
  running the command. Below are few examples of how it works:

    shell_call(['cp', 'a', 'b'], a='asd') calls command 'cp a b'

    shell_call(['cp', '${a}', 'b'], a='asd') calls command 'cp asd b',
    '${a}; was replaced with 'asd' before calling the command
  """
  # Regular expression to find instances of '${NAME}' in a string
  CMD_VARIABLE_RE = re.compile('^\\$\\{(\\w+)\\}$')
  command = list(command)
  for i in range(len(command)):
    m = CMD_VARIABLE_RE.match(command[i])
    if m:
      var_id = m.group(1)
      if var_id in kwargs:
        command[i] = kwargs[var_id]
  str_command = ' '.join(command)
  logging.debug('Executing shell command: %s' % str_command)
  return subprocess.check_output(command)

def deep_copy(numpy_dict):
  """
  Returns a copy of a dictionary whose values are numpy arrays.
  Copies their values rather than copying references to them.
  """
  out = {}
  for key in numpy_dict:
    out[key] = numpy_dict[key].copy()
  return out



import os
import warnings

import cleverhans


def check_installation(cur_file):
  # Warn user if running cleverhans from a different directory than tutorial.
  cur_dir = os.path.split(os.path.dirname(os.path.abspath(cur_file)))[0]
  ch_dir = os.path.split(cleverhans.__path__[0])[0]
  if cur_dir != ch_dir:
    warnings.warn("It appears that you have at least two versions of "
                  "cleverhans installed, one at %s and one at"
                  " %s. You are running the tutorial script from the "
                  "former but python imported the library module from the "
                  "latter. This may cause errors, for example if the tutorial"
                  " version is newer than the library version and attempts to"
                  " call new features." % (cur_dir, ch_dir))



"""Code for running the certification problem."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf

from cleverhans.experimental.certification import dual_formulation
from cleverhans.experimental.certification import neural_net_params
from cleverhans.experimental.certification import optimization
from cleverhans.experimental.certification import read_weights
from cleverhans.experimental.certification import utils


flags = tf.app.flags
FLAGS = flags.FLAGS
flags.DEFINE_string('checkpoint', None,
                    'Path of checkpoint with trained model to verify')
flags.DEFINE_string('model_json', None,
                    'Path of json file with model description')
flags.DEFINE_string('init_dual_file', None,
                    'Path of numpy file with dual variables to initialize')
flags.DEFINE_string('test_input', None,
                    'Path of numpy file with test input to certify')
flags.DEFINE_integer('true_class', 0,
                     'True class of the test input')
flags.DEFINE_integer('adv_class', -1,
                     'target class of adversarial example; all classes if -1')
flags.DEFINE_float('input_minval', -1,
                   'Minimum value of valid input')
flags.DEFINE_float('input_maxval', 1,
                   'Maximum value of valid input')
flags.DEFINE_float('epsilon', 0.2,
                   'Size of perturbation')
# Nu might need tuning based on the network
flags.DEFINE_float('init_nu', 300.0,
                   'Initialization of nu variable.')
flags.DEFINE_float('init_penalty', 100.0,
                   'Initial penalty')
flags.DEFINE_integer('small_eig_num_steps', 500,
                     'Number of eigen value steps in intermediate iterations')
flags.DEFINE_integer('large_eig_num_steps', 5000,
                     'Number of eigen value steps in each outer iteration')
flags.DEFINE_integer('inner_num_steps', 600,
                     'Number of steps to run in inner loop')
flags.DEFINE_float('outer_num_steps', 10,
                   'Number of steps to run in outer loop')
flags.DEFINE_float('beta', 2,
                   'Multiplicative factor to increase penalty by')
flags.DEFINE_float('smoothness_parameter', 0.001,
                   'Smoothness parameter if using eigen decomposition')
flags.DEFINE_float('eig_learning_rate', 0.001,
                   'Learning rate for computing min eigen value')
flags.DEFINE_string('optimizer', 'adam',
                    'Optimizer to use for entire optimization')
flags.DEFINE_float('init_learning_rate', 0.1,
                   'Initial learning rate')
flags.DEFINE_float('learning_rate_decay', 0.1,
                   'Decay of learning rate')
flags.DEFINE_float('momentum_parameter', 0.9,
                   'Momentum parameter if using momentum optimizer')
flags.DEFINE_integer('print_stats_steps', 50,
                     'Number of steps to print stats after')
flags.DEFINE_string('stats_folder', None,
                    'Folder to save stats of the iterations')
flags.DEFINE_integer('projection_steps', 200,
                     'Number of steps to compute projection after')
flags.DEFINE_integer('num_classes', 10,
                     'Total number of classes')


def main(_):
  net_weights, net_biases, net_layer_types = read_weights.read_weights(
      FLAGS.checkpoint, FLAGS.model_json)
  nn_params = neural_net_params.NeuralNetParams(
      net_weights, net_biases, net_layer_types)
  print(nn_params.sizes)
  dual_var = utils.initialize_dual(nn_params, FLAGS.init_dual_file,
                                   init_nu=FLAGS.init_nu)
  # Reading test input and reshaping
  with tf.gfile.Open(FLAGS.test_input) as f:
    test_input = np.load(f)
  test_input = np.reshape(test_input, [np.size(test_input), 1])

  if FLAGS.adv_class == -1:
    start_class = 0
    end_class = FLAGS.num_classes
  else:
    start_class = FLAGS.adv_class
    end_class = FLAGS.adv_class + 1
  for adv_class in range(start_class, end_class):
    print('Adv class', adv_class)
    if adv_class == FLAGS.true_class:
      continue
    dual = dual_formulation.DualFormulation(dual_var,
                                            nn_params,
                                            test_input,
                                            FLAGS.true_class,
                                            adv_class,
                                            FLAGS.input_minval,
                                            FLAGS.input_maxval,
                                            FLAGS.epsilon)
    dual.set_differentiable_objective()
    dual.get_full_psd_matrix()
    optimization_params = {
        'init_penalty': FLAGS.init_penalty,
        'large_eig_num_steps': FLAGS.large_eig_num_steps,
        'small_eig_num_steps': FLAGS.small_eig_num_steps,
        'inner_num_steps': FLAGS.inner_num_steps,
        'outer_num_steps': FLAGS.outer_num_steps,
        'beta': FLAGS.beta,
        'smoothness_parameter': FLAGS.smoothness_parameter,
        'eig_learning_rate': FLAGS.eig_learning_rate,
        'optimizer': FLAGS.optimizer,
        'init_learning_rate': FLAGS.init_learning_rate,
        'learning_rate_decay': FLAGS.learning_rate_decay,
        'momentum_parameter': FLAGS.momentum_parameter,
        'print_stats_steps': FLAGS.print_stats_steps,
        'stats_folder': FLAGS.stats_folder,
        'projection_steps': FLAGS.projection_steps}
    with tf.Session() as sess:
      sess.run(tf.global_variables_initializer())
      optimization_object = optimization.Optimization(dual,
                                                      sess,
                                                      optimization_params)
      optimization_object.prepare_one_step()
      is_cert_found = optimization_object.run_optimization()
      if not is_cert_found:
        print('Current example could not be verified')
        exit()
  print('Example successfully verified')

if __name__ == '__main__':
  tf.app.run(main)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import tempfile
import warnings

from cleverhans import dataset

utils_mnist_warning = "cleverhans.utils_mnist is deprecrated and will be " \
                      "removed on or after 2019-03-26. Switch to " \
                      "cleverhans.dataset instead."


def maybe_download_mnist_file(file_name, datadir=None, force=False):
  warnings.warn(utils_mnist_warning)
  url = os.path.join('http://yann.lecun.com/exdb/mnist/', file_name)
  return dataset.maybe_download_file(url, datadir=None, force=False)


def download_and_parse_mnist_file(file_name, datadir=None, force=False):
  warnings.warn(utils_mnist_warning)
  return dataset.download_and_parse_mnist_file(file_name, datadir=None,
                                               force=False)


def data_mnist(datadir=tempfile.gettempdir(), train_start=0,
               train_end=60000, test_start=0, test_end=10000):
  warnings.warn(utils_mnist_warning)
  mnist = dataset.MNIST(train_start=train_start,
                        train_end=train_end,
                        test_start=test_start,
                        test_end=test_end,
                        center=False)
  return mnist.get_set('train') + mnist.get_set('test')



"""
Model construction utilities based on keras
"""
from distutils.version import LooseVersion
import keras
from keras.models import Sequential
from keras.layers import Dense, Activation, Flatten

from .model import Model, NoSuchLayerError

if LooseVersion(keras.__version__) >= LooseVersion('2.0.0'):
  from keras.layers import Conv2D
else:
  from keras.layers import Convolution2D


def conv_2d(filters, kernel_shape, strides, padding, input_shape=None):
  """
  Defines the right convolutional layer according to the
  version of Keras that is installed.
  :param filters: (required integer) the dimensionality of the output
                  space (i.e. the number output of filters in the
                  convolution)
  :param kernel_shape: (required tuple or list of 2 integers) specifies
                       the strides of the convolution along the width and
                       height.
  :param padding: (required string) can be either 'valid' (no padding around
                  input or feature map) or 'same' (pad to ensure that the
                  output feature map size is identical to the layer input)
  :param input_shape: (optional) give input shape if this is the first
                      layer of the model
  :return: the Keras layer
  """
  if LooseVersion(keras.__version__) >= LooseVersion('2.0.0'):
    if input_shape is not None:
      return Conv2D(filters=filters, kernel_size=kernel_shape,
                    strides=strides, padding=padding,
                    input_shape=input_shape)
    else:
      return Conv2D(filters=filters, kernel_size=kernel_shape,
                    strides=strides, padding=padding)
  else:
    if input_shape is not None:
      return Convolution2D(filters, kernel_shape[0], kernel_shape[1],
                           subsample=strides, border_mode=padding,
                           input_shape=input_shape)
    else:
      return Convolution2D(filters, kernel_shape[0], kernel_shape[1],
                           subsample=strides, border_mode=padding)


def cnn_model(logits=False, input_ph=None, img_rows=28, img_cols=28,
              channels=1, nb_filters=64, nb_classes=10):
  """
  Defines a CNN model using Keras sequential model
  :param logits: If set to False, returns a Keras model, otherwise will also
                  return logits tensor
  :param input_ph: The TensorFlow tensor for the input
                  (needed if returning logits)
                  ("ph" stands for placeholder but it need not actually be a
                  placeholder)
  :param img_rows: number of row in the image
  :param img_cols: number of columns in the image
  :param channels: number of color channels (e.g., 1 for MNIST)
  :param nb_filters: number of convolutional filters per layer
  :param nb_classes: the number of output classes
  :return:
  """
  model = Sequential()

  # Define the layers successively (convolution layers are version dependent)
  if keras.backend.image_dim_ordering() == 'th':
    input_shape = (channels, img_rows, img_cols)
  else:
    input_shape = (img_rows, img_cols, channels)

  layers = [conv_2d(nb_filters, (8, 8), (2, 2), "same",
                    input_shape=input_shape),
            Activation('relu'),
            conv_2d((nb_filters * 2), (6, 6), (2, 2), "valid"),
            Activation('relu'),
            conv_2d((nb_filters * 2), (5, 5), (1, 1), "valid"),
            Activation('relu'),
            Flatten(),
            Dense(nb_classes)]

  for layer in layers:
    model.add(layer)

  if logits:
    logits_tensor = model(input_ph)
  model.add(Activation('softmax'))

  if logits:
    return model, logits_tensor
  else:
    return model


class KerasModelWrapper(Model):
  """
  An implementation of `Model` that wraps a Keras model. It
  specifically exposes the hidden features of a model by creating new models.
  The symbolic graph is reused and so there is little overhead. Splitting
  in-place operations can incur an overhead.
  """

  def __init__(self, model):
    """
    Create a wrapper for a Keras model
    :param model: A Keras model
    """
    super(KerasModelWrapper, self).__init__(None, None, {})

    if model is None:
      raise ValueError('model argument must be supplied.')

    self.model = model
    self.keras_model = None

  def _get_softmax_name(self):
    """
    Looks for the name of the softmax layer.
    :return: Softmax layer name
    """
    for layer in self.model.layers:
      cfg = layer.get_config()
      if 'activation' in cfg and cfg['activation'] == 'softmax':
        return layer.name

    raise Exception("No softmax layers found")

  def _get_logits_name(self):
    """
    Looks for the name of the layer producing the logits.
    :return: name of layer producing the logits
    """
    softmax_name = self._get_softmax_name()
    softmax_layer = self.model.get_layer(softmax_name)

    if not isinstance(softmax_layer, Activation):
      # In this case, the activation is part of another layer
      return softmax_name

    if not hasattr(softmax_layer, '_inbound_nodes'):
      raise RuntimeError("Please update keras to version >= 2.1.3")

    node = softmax_layer._inbound_nodes[0]

    logits_name = node.inbound_layers[0].name

    return logits_name

  def get_logits(self, x):
    """
    :param x: A symbolic representation of the network input.
    :return: A symbolic representation of the logits
    """
    logits_name = self._get_logits_name()
    logits_layer = self.get_layer(x, logits_name)

    # Need to deal with the case where softmax is part of the
    # logits layer
    if logits_name == self._get_softmax_name():
      softmax_logit_layer = self.get_layer(x, logits_name)

      # The final op is the softmax. Return its input
      logits_layer = softmax_logit_layer._op.inputs[0]

    return logits_layer

  def get_probs(self, x):
    """
    :param x: A symbolic representation of the network input.
    :return: A symbolic representation of the probs
    """
    name = self._get_softmax_name()

    return self.get_layer(x, name)

  def get_layer_names(self):
    """
    :return: Names of all the layers kept by Keras
    """
    layer_names = [x.name for x in self.model.layers]
    return layer_names

  def fprop(self, x):
    """
    Exposes all the layers of the model returned by get_layer_names.
    :param x: A symbolic representation of the network input
    :return: A dictionary mapping layer names to the symbolic
             representation of their output.
    """
    from keras.models import Model as KerasModel

    if self.keras_model is None:
      # Get the input layer
      new_input = self.model.get_input_at(0)

      # Make a new model that returns each of the layers as output
      out_layers = [x_layer.output for x_layer in self.model.layers]
      self.keras_model = KerasModel(new_input, out_layers)

    # and get the outputs for that model on the input x
    outputs = self.keras_model(x)

    # Keras only returns a list for outputs of length >= 1, if the model
    # is only one layer, wrap a list
    if len(self.model.layers) == 1:
      outputs = [outputs]

    # compute the dict to return
    fprop_dict = dict(zip(self.get_layer_names(), outputs))

    return fprop_dict

  def get_layer(self, x, layer):
    """
    Expose the hidden features of a model given a layer name.
    :param x: A symbolic representation of the network input
    :param layer: The name of the hidden layer to return features at.
    :return: A symbolic representation of the hidden features
    :raise: NoSuchLayerError if `layer` is not in the model.
    """
    # Return the symbolic representation for this layer.
    output = self.fprop(x)
    try:
      requested = output[layer]
    except KeyError:
      raise NoSuchLayerError()
    return requested



"""Dataset class for CleverHans

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import array
import functools
import gzip
import operator
import os
import struct
import tempfile
import sys
import numpy as np
import tensorflow as tf

from cleverhans import utils

# Only load keras if user tries to use a dataset that requires it
keras_imported = False


class Dataset(object):
  """Abstract base class representing a dataset.
  """

  # The number of classes in the dataset. Should be specified by subclasses.
  NB_CLASSES = None

  def __init__(self, kwargs=None):
    if kwargs is None:
      kwargs = {}
    if "self" in kwargs:
      del kwargs["self"]
    self.kwargs = kwargs

  def get_factory(self):
    """Returns a picklable callable that recreates the dataset.
    """

    return Factory(type(self), self.kwargs)

  def get_set(self, which_set):
    """Returns the training set or test set as an (x_data, y_data) tuple.
    :param which_set: 'train' or 'test'
    """
    return (getattr(self, 'x_' + which_set),
            getattr(self, 'y_' + which_set))

  def to_tensorflow(self):
    raise NotImplementedError()

  @classmethod
  def in_memory_dataset(cls, x, y, shuffle=None, repeat=True):
    assert x.shape[0] == y.shape[0]
    d = tf.data.Dataset.range(x.shape[0])
    if repeat:
      d = d.repeat()
    if shuffle:
      d = d.shuffle(shuffle)

    def lookup(p):
      return x[p], y[p]
    d = d.map(lambda i: tf.py_func(lookup, [i], [tf.float32] * 2))
    return d


class MNIST(Dataset):
  """The MNIST dataset"""

  NB_CLASSES = 10

  def __init__(self, train_start=0, train_end=60000, test_start=0,
               test_end=10000, center=False, max_val=1.):
    super(MNIST, self).__init__(locals())
    x_train, y_train, x_test, y_test = data_mnist(train_start=train_start,
                                                  train_end=train_end,
                                                  test_start=test_start,
                                                  test_end=test_end)

    if center:
      x_train = x_train * 2. - 1.
      x_test = x_test * 2. - 1.
    x_train *= max_val
    x_test *= max_val

    self.x_train = x_train.astype('float32')
    self.y_train = y_train.astype('float32')
    self.x_test = x_test.astype('float32')
    self.y_test = y_test.astype('float32')

  def to_tensorflow(self, shuffle=4096):
    return (self.in_memory_dataset(self.x_train, self.y_train, shuffle),
            self.in_memory_dataset(self.x_test, self.y_test, repeat=False))


class CIFAR10(Dataset):
  """The CIFAR-10 dataset"""

  NB_CLASSES = 10

  LABEL_NAMES = ["airplane", "automobile", "bird", "cat", "deer", "dog",
                 "frog", "horse", "ship", "truck"]

  def __init__(self, train_start=0, train_end=60000, test_start=0,
               test_end=10000, center=False, max_val=1.):
    super(CIFAR10, self).__init__(locals())
    packed = data_cifar10(train_start=train_start,
                          train_end=train_end,
                          test_start=test_start,
                          test_end=test_end)
    x_train, y_train, x_test, y_test = packed

    if center:
      x_train = x_train * 2. - 1.
      x_test = x_test * 2. - 1.
    x_train *= max_val
    x_test *= max_val

    self.x_train = x_train
    self.y_train = y_train
    self.x_test = x_test
    self.y_test = y_test
    self.max_val = max_val

  def to_tensorflow(self, shuffle=4096):
    # This is much more efficient with data augmentation, see tutorials.
    return (self.in_memory_dataset(self.x_train, self.y_train, shuffle),
            self.in_memory_dataset(self.x_test, self.y_test, repeat=False))


class Factory(object):
  """
  A callable that creates an object of the specified type and configuration.
  """

  def __init__(self, cls, kwargs):
    self.cls = cls
    self.kwargs = kwargs

  def __call__(self):
    """Returns the created object.
    """
    return self.cls(**self.kwargs)


def maybe_download_file(url, datadir=None, force=False):
  try:
    from urllib.request import urlretrieve
  except ImportError:
    from urllib import urlretrieve

  if not datadir:
    datadir = tempfile.gettempdir()
  file_name = url[url.rfind("/")+1:]
  dest_file = os.path.join(datadir, file_name)

  isfile = os.path.isfile(dest_file)

  if force or not isfile:
    urlretrieve(url, dest_file)
  return dest_file


def download_and_parse_mnist_file(file_name, datadir=None, force=False):
  url = os.path.join('http://yann.lecun.com/exdb/mnist/', file_name)
  file_name = maybe_download_file(url, datadir=datadir, force=force)

  # Open the file and unzip it if necessary
  if os.path.splitext(file_name)[1] == '.gz':
    open_fn = gzip.open
  else:
    open_fn = open

  # Parse the file
  with open_fn(file_name, 'rb') as file_descriptor:
    header = file_descriptor.read(4)
    assert len(header) == 4

    zeros, data_type, n_dims = struct.unpack('>HBB', header)
    assert zeros == 0

    hex_to_data_type = {
        0x08: 'B',
        0x09: 'b',
        0x0b: 'h',
        0x0c: 'i',
        0x0d: 'f',
        0x0e: 'd'}
    data_type = hex_to_data_type[data_type]

    # data_type unicode to ascii conversion (Python2 fix)
    if sys.version_info[0] < 3:
      data_type = data_type.encode('ascii', 'ignore')

    dim_sizes = struct.unpack(
        '>' + 'I' * n_dims,
        file_descriptor.read(4 * n_dims))

    data = array.array(data_type, file_descriptor.read())
    data.byteswap()

    desired_items = functools.reduce(operator.mul, dim_sizes)
    assert len(data) == desired_items
    return np.array(data).reshape(dim_sizes)


def data_mnist(datadir=tempfile.gettempdir(), train_start=0,
               train_end=60000, test_start=0, test_end=10000):
  """
  Load and preprocess MNIST dataset
  :param datadir: path to folder where data should be stored
  :param train_start: index of first training set example
  :param train_end: index of last training set example
  :param test_start: index of first test set example
  :param test_end: index of last test set example
  :return: tuple of four arrays containing training data, training labels,
           testing data and testing labels.
  """
  assert isinstance(train_start, int)
  assert isinstance(train_end, int)
  assert isinstance(test_start, int)
  assert isinstance(test_end, int)

  X_train = download_and_parse_mnist_file(
      'train-images-idx3-ubyte.gz', datadir=datadir) / 255.
  Y_train = download_and_parse_mnist_file(
      'train-labels-idx1-ubyte.gz', datadir=datadir)
  X_test = download_and_parse_mnist_file(
      't10k-images-idx3-ubyte.gz', datadir=datadir) / 255.
  Y_test = download_and_parse_mnist_file(
      't10k-labels-idx1-ubyte.gz', datadir=datadir)

  X_train = np.expand_dims(X_train, -1)
  X_test = np.expand_dims(X_test, -1)

  X_train = X_train[train_start:train_end]
  Y_train = Y_train[train_start:train_end]
  X_test = X_test[test_start:test_end]
  Y_test = Y_test[test_start:test_end]

  Y_train = utils.to_categorical(Y_train, nb_classes=10)
  Y_test = utils.to_categorical(Y_test, nb_classes=10)
  return X_train, Y_train, X_test, Y_test


def data_cifar10(train_start=0, train_end=50000, test_start=0, test_end=10000):
  """
  Preprocess CIFAR10 dataset
  :return:
  """

  global keras_imported
  if not keras_imported:
    import keras
    from keras.datasets import cifar10
    from keras.utils import np_utils

  # These values are specific to CIFAR10
  img_rows = 32
  img_cols = 32
  nb_classes = 10

  # the data, shuffled and split between train and test sets
  (x_train, y_train), (x_test, y_test) = cifar10.load_data()

  if keras.backend.image_dim_ordering() == 'th':
    x_train = x_train.reshape(x_train.shape[0], 3, img_rows, img_cols)
    x_test = x_test.reshape(x_test.shape[0], 3, img_rows, img_cols)
  else:
    x_train = x_train.reshape(x_train.shape[0], img_rows, img_cols, 3)
    x_test = x_test.reshape(x_test.shape[0], img_rows, img_cols, 3)
  x_train = x_train.astype('float32')
  x_test = x_test.astype('float32')
  x_train /= 255
  x_test /= 255
  print('x_train shape:', x_train.shape)
  print(x_train.shape[0], 'train samples')
  print(x_test.shape[0], 'test samples')

  # convert class vectors to binary class matrices
  y_train = np_utils.to_categorical(y_train, nb_classes)
  y_test = np_utils.to_categorical(y_test, nb_classes)

  x_train = x_train[train_start:train_end, :, :, :]
  y_train = y_train[train_start:train_end, :]
  x_test = x_test[test_start:test_end, :]
  y_test = y_test[test_start:test_end, :]

  return x_train, y_train, x_test, y_test
